/*
Copyright (c) 1996-2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.jcpufreqd;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;


/**Simple file utilities
 */
public final class FileTools
    {
    private FileTools() { } // Prevent creation of an instance.


    /**8-bit ISO-8859-1 file encoding. */
    public static final String FILE_ENCODING_8859_1 = "ISO-8859-1";

    /**Read text file into List of String lines; never null.
     * Reads a line at a time, trimming whitespace off either end.
     * <p>
     * This treats the file as ISO-8859-1 8-bit data.
     *
     * @exception IOException  in case of trouble
     */
    public static List<String> readTextFileLines(final File f)
        throws IOException
        {
        final List<String> result = new ArrayList<String>();
        BufferedReader br = null;
        try {
            br = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream(f), FILE_ENCODING_8859_1));
            String inputLine;
            while((inputLine = br.readLine()) != null)
                { result.add(inputLine.trim()); }
            }
        finally
            {
            // Try to close the file (possibly provoking an exception)...
            if(br != null) { br.close(); br = null; }
            }
        return(result);
        }

    /**Write (from offset 0) exactly the given text to the given extant output file.
     * No attempt is made to append whitespace or line terminators, etc.
     * <p>
     * The text is treated as is ISO-8859-1 text,
     * any should not contain any non-8-bit values,
     * and indeed should be printable 7-bit ASCII in most normal use with Linux sysfs.
     * <p>
     * This aborts if the file does not already exist,
     * ie avoids trying to create a new file.
     */
    public static void writeTextFileInPlace(final File f, final String s)
        throws IOException
        {
        if(!f.exists()) { throw new FileNotFoundException("will not create file: " + f); }
        final Writer w = new OutputStreamWriter(new FileOutputStream(f, false), FILE_ENCODING_8859_1);
        try { w.write(s);  }
        // Release resources.
        finally { w.close(); }
        }
    }
