/*
Copyright (c) 2007, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.jcpufreqd;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.security.SecureRandom;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.regex.Pattern;


/**Lightweight program to help CPU power management on Linux.
 * This is a simplified/modified replacement for cpufreqd,
 * with some features to assist power management in a system running from
 * scarce RE (Renewable Energy) such as solar PV.
 * <p>
 * In particular this switches to a lower-power energy-conserving mode
 * whenever there is not abundant external power.
 * Even 'external' power (seen by ACPI, etc, as 'AC' power)
 * may not be coming from mains AC, but from external battery supply.
 * <p>
 * This can also be used to force the lower-power mode
 * if the CPU gets too hot as high temperature reduces system lifetime
 * and represents a waste of energy (and the fan noise may be irritating).
 * <p>
 * This expects a kernel-based CPU frequency governor such as 'ondemand'
 * to do most of any real-time work required,
 * with this merely slowly modulating the parameters.
 * This may therefore poll slowly to minimise its own wakeups and CPU time,
 * possibly using events to wake up for quicker response to some changes.
 * <p>
 * This uses the Linux sysfs (/sys) interfaces,
 * and will need to be run as root to modify any parameters/settings.
 * <p>
 * This attempts to be robust in the face of CPUs going on or off-line,
 * and may attempt to manage CPUs' frequencies, etc, separately.
 * <p>
 * This attempts to be 'fail-safe' by reverting to low-power mode
 * if it encounters a condition or error that it does not understand.
 */
public final class JCPUFreqD
    {
    /**Run the program from the command-line, eg as a daemon from init.
     * @param args  the command-line arguments; never null
     */
    public static void main(String[] args)
        {
        final JCPUFreqD instance = new JCPUFreqD();

        // We can suppress the lockfile if we really want to.
        final boolean suppressLock = Boolean.getBoolean("org.hd.d.JCPUFreqD.suppressLock");
        // If the lock is being used, atomically create it now or exit.
        final long lockUID = suppressLock ? 0 : ((new SecureRandom()).nextLong() >>> 1);
        final File lockFile = suppressLock ? null : DEFAULT_RUN_LOCK_FILE;
        if(!suppressLock)
            {
            try
                {
                if(!lockFile.createNewFile())
                    {
                    errLogger .log("Lock file exists, cannot start: " + lockFile);
                    System.exit(1);
                    }
                // Write out unique value into the lock file.
                FileTools.writeTextFileInPlace(lockFile, String.valueOf(lockUID));
                }
            catch(final Exception e)
                {
                errLogger.log("Cannot create lock file: " + lockFile);
                System.exit(1);
                }            
            }
        else { errLogger.log("NOTE: lock file suppressed"); }

        // Start the main loop.  Never exits.
        instance.mainLoop(lockFile,
                          lockUID);
        }

    /**Simple static logger for non-critical info. */
    static final SimpleLoggerIF logger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            {
            // Log to the Java console.
            System.out.println(message);
            }
        };

    /**Simple static logger for errors. */
    static final SimpleLoggerIF errLogger = new SimpleLoggerIF()
        {
        public final void log(final String message)
            {
            // Log to the Java console.
            System.err.println(message);
            }
        };

    /**Main loop to check the system/power state and set parameters, then sleep, rinse, repeat.
     * All the instance parameters should have been set by now.
     * <p>
     * Never exits unless our lockfile goes missing or is replaced.
     *
     * @param lockFile  if non-null this is the exclusive lockfile for this process
     * @param uid  if lockFile is non-null this is the unique ID that should be in our lockfile
     */
    void mainLoop(final File lockFile, final long uid)
        {
        if(debug) { logger.log("jCPUFreqD starting..."); }

        final String lockUID = String.valueOf(uid);
        if(debug && (lockFile != null)) { logger.log("Unique ID: " + uid); }

        if(debug) { logger.log("Initial CPUs available: " + Arrays.toString(enumerateCPUs())); }

        // Main loop.
        for( ; ; )
            {
            // Check that we still hold the lock (if there is one); quit otherwise.
            if(lockFile != null)
                {
                try
                    {
                    final List<String> lockFileData = FileTools.readTextFileLines(lockFile);
                    if((lockFileData.size() != 1) || !lockUID.equals(lockFileData.get(0)))
                        {
                        errLogger.log("Lockfile replaced, quitting: " + lockFile);
                        System.exit(1); // Error exit.
                        }
                    }
                catch(final Exception e)
                    {
                    errLogger.log("Lockfile removed or unreadable, quitting: " + lockFile);
                    System.exit(1); // Error exit.
                    }
                }

            // Default to a low-power state for safety.
            boolean lowPower = true;

            try
                {
                // Compute target power state.
                // Stop (break) as soon as we find any reason to stay in the low-power state
                // and/or when we encounter an unexpected serious error.
                checkInputs: do
                    {
                    // If any of the low-power flag files is present
                    // then stop looking for anything else
                    // and leave us in low-power mode.
                    // Do this first because it's probably the 'cheapest' test.
                    for(final File f : lowPowerFlagFiles)
                        {
                        if(f.exists())
                            {
                            if(debug) { logger.log("Low-power flag file present: " + f); }
                            assert lowPower : "Must be in low-power mode if flag file present.";
                            break checkInputs; // Break from outer loop too...
                            }
                        }
                    if(debug) { logger.log("No low-power flag files present."); }

                    // Confirm that we're still on external power.
                    // If we're on battery power (AC power is 'off-line')
                    // then we must stay in low-power mode.
                    // If we cannot read the correct status files,
                    // fallback to low-power mode.
                    final List<String> acPowerState = FileTools.readTextFileLines(PROC_AC_POWER_STATUS_FILE);
                    // We expect exactly on line of the form:
                    // <code>status: on-line</code> (with external power available)
                    // or <code>status: off-line</code> (when the machine is running on internal batteries).
                    if(acPowerState.size() != 1)
                        {
                        errLogger.log("expected exactly one status line in " + PROC_AC_POWER_STATUS_FILE);
                        break;
                        }
                    final String words[] = acPowerState.get(0).split(" +");
                    if(words.length != 2)
                        {
                        errLogger.log("expected exactly two words in " + PROC_AC_POWER_STATUS_FILE);
                        break;
                        }
                    if(!"on-line".equals(words[1]))
                        {
                        // No external power, so must go into low-power mode.
                        if(debug) { logger.log("No external power, so must go into low-power mode."); }
                        assert(lowPower);
                        break;
                        }
                    if(debug) { logger.log("'AC' external power on-line."); }



                    // TODO: remaining tests, eg for CPU temperature.



                    // OK, no problem or reason to stay low-power found,
                    // so we can use high-power mode.
                    lowPower = false;
                    } while(false);
                }
            // In the case of an unexpected exception,
            // make sure that we force low-power mode.
            // We avoid aborting the main loop.
            catch(final Throwable e)
                {
                lowPower = true;
                errLogger.log("Unexpected exception, forcing low-power mode: " + e.getMessage());
                }            
            if(debug) { logger.log("Target LOW power state: " + lowPower); }

            // Now enforce the selected mode.
            setMode(lowPower);

            // Sleep longer before the next poll when in low-power mode,
            // ie be conservative about going back into full-power mode
            // and reduce also resources spent on the polling work itself.
            final long basicSleepTime = 1000 * poll_s;
            try{ Thread.sleep(lowPower ? (basicSleepTime << 2) : basicSleepTime); }
            catch(final InterruptedException e)
                {
                errLogger.log("Interrupted: " + e.getMessage());
                System.exit(1); // Exit with an error.
                }            
            }
        }

    /**Default (immutable) lock file to prevent multiple instances running at once; never null.
     * This is created atomically, and written with a locally-generated unique number, partially random.
     * <p>
     * The file is checked periodically, and if it is missing then we shut down:
     * so to shut down the process, remove the file (and wait a little while).
     * <p>
     * When the file is checked it is re-read, and if it contains a different value to the one that we wrote,
     * then we shut down, on the grounds that a new daemon must have started
     * (possibly immediately after a previous lock file was removed).
     * <p>
     * The aim is to ensure that multiple instances do not run concurrently,
     * at least not for very long, to avoid conflict.
     */
    public static final File DEFAULT_RUN_LOCK_FILE =
        new File("/var/run/JCPUFreqD.lock");

    /**Default (immutable) flag file set by us to indicate low-power mode; null if no such output flag file.
     * We create the file on entering low-power mode,
     * and remove it on existing low-power mode,
     * where possible.
     */
    public static final File DEFAULT_LOW_POWER_OUTPUT_FLAG_FILE =
        new File("/var/run/RUNNING_SLOW.flag");

    /**Flag file set by us to indicate low-power mode; null if no such output flag file.
     * We create the file on entering low-power mode,
     * and remove it on existing low-power mode,
     * where possible.
     */
    public final File outputFlagFile = DEFAULT_LOW_POWER_OUTPUT_FLAG_FILE;

    /**Base directory in sysfs /sys filesystem for CPU information and controls; never null. */
    public static final File SYSFS_base_CPU_dir = new File("/sys/devices/system/cpu");

    /**Sysfs flag file to set to 1 to minimise number of cores carrying processing load; never null.
     * Set the value to 0 for normal load balancing.
     */
    public static final File SYSFS_sched_mc_power_savings = new File(SYSFS_base_CPU_dir, "sched_mc_power_savings");

    /**Compiled regex for matching CPU dirs; never null. */
    private static final Pattern cpuDirRegex = Pattern.compile("cpu[0-9]+");

    /**Computes a sorted list of visible CPUs under the sysfs CPU base area; never null but may be empty.
     * This looks for directories called 'cpu<i>NN</i>' where N is a non-negative integer,
     * and returns them in lexically-sorted order,
     * which should at least ensure that the lowest-numbered CPU
     * (which may well end up handling most interrupts, and more work with sched_mc_power_savings)
     * is first in the result.
     * This first value can then be handled specially,
     * eg given a wider range of allowed CPU speeds than the others.
     */
    private static String[] enumerateCPUs()
        {
        final String[] result = SYSFS_base_CPU_dir.list(new FilenameFilter() {
            /**Accept directories of form cpuNN. */
            public boolean accept(final File dir, final String name)
                {
                // Quickly reject names with the wrong syntax.
                if(!cpuDirRegex.matcher(name).matches()) { return(false); }
                // As a check, reject entries that are not readable directories for any reason.
                final File fullname = new File(dir, name);
                if(!fullname.isDirectory() || !fullname.canRead()) { return(false); }
                // Seems OK.
                return(true);
                }
            });
        if(result == null) { return(new String[0]); }
        Arrays.sort(result);
        return(result);
        }

    /**Set the appropriate power mode.
     * When in low-power mode we:
     * <ul>
     * <li>TODO: Set all (independent) cores to use the 'ondemand' governor with
     *     the lowest-numbered one having a frequency range from minimum to about half maximum
     *     and all the other cores restricted to minimum if possible.
     * <li>Set the 'sched_mc_power_savings' flag to do more processing on one core if possible.
     * <li>Create the low-power output flag file if possible (and not already present).
     * </ul>
     * else, in full-power mode, we:
     * <ul>
     * <li>TODO: Set all (independent) cores to use the 'ondemand' governor with the full frequency range.
     * <li>Clear the 'sched_mc_power_savings' flag to spread processing over all cores if possible.
     * <li>Remove the low-power output flag file if possible (and if present).
     * </ul>
     *
     * @param lowPower  if true then put the machine into a power-conserving mode
     */
    private void setMode(final boolean lowPower)
        {
        // Apply appropriate settings to each CPU.
        // TODO: this currently assumes that all CPUs here are independent for P and C states.
        final String[] cpus = enumerateCPUs();
        // For each CPU try to:
        //   * Set the right governor and associated parameters (if possible).
        //   * Set the right clockspeed range.
        for(final String cpu : cpus)
            {
            // Compute this CPU's base dir.
            final File cpuDir = new File(SYSFS_base_CPU_dir, cpu);
            // Compute the cpufreq subdir which is where everything of interest lives.
            final File cpufreqDir = new File(cpuDir, "cpufreq");
            if(!cpufreqDir.isDirectory())
                {
                errLogger.log("Cannot access cpufreq dir to adjust CPU frequency (make sure kernel cpufreq modules are loaded): " + cpufreqDir);
                break; // Assume other CPUs have no cpufreq access too.
                }

            // Set the preferred governor unconditionally.
            try { FileTools.writeTextFileInPlace(new File(cpufreqDir, "scaling_governor"), freqGovernor); }
            catch(final IOException e) { errLogger.log("Could not set governor for "+cpu+": " + e.getMessage()); }

            // Attempt to set the min/max CPU speed bounds, regardless of governor.
            // (Even if we can't set the governor to our preferred one, this may help.)
            // For full-power mode this is always the full range (assumed to be the first and last values).
            // For low-power mode this is always the lowest speed allowed (assumed to be the last value)
            // except for the first CPU, where the range is a middle value down to the lowest.
            try
                {
                final File scaling_available_frequencies = new File(cpufreqDir, "scaling_available_frequencies");
                final List<String> availableSpeeds = FileTools.readTextFileLines(scaling_available_frequencies);
                if(availableSpeeds.size() < 1) { errLogger.log("Could not read scaling_available_frequencies file for "+cpu+": "+scaling_available_frequencies); }
                else
                    {
                    final String[] speeds = availableSpeeds.get(0).split(" +");
                    if(speeds.length < 1) { errLogger.log("Could not read speeds from scaling_available_frequencies file for "+cpu); }
                    else
                        {
                        // Minimum speed is always the lowest (last) one listed.
                        final String minSpeed = speeds[speeds.length-1];
                        // Maximum speed depends on the target power mode and which CPU if more than one.
                        final String maxSpeed = (!lowPower) ? speeds[0] :
                            (cpu.equals(cpus[0]) ? speeds[speeds.length/2] : minSpeed);
                        if(debug) { logger.log("Selected max/min speeds for "+cpu+": "+maxSpeed+"/"+minSpeed); }
                        FileTools.writeTextFileInPlace(new File(cpufreqDir, "scaling_min_freq"), minSpeed);
                        FileTools.writeTextFileInPlace(new File(cpufreqDir, "scaling_max_freq"), maxSpeed);
                        }
                    }
                }
            catch(final Exception e) { errLogger.log("Could not set sampling_rate flag for "+cpu+": " + e.getMessage()); }            

            // If the preferred governer is 'ondemand' and is available then tweak some parameters.
            if("ondemand".equals(freqGovernor))
                {
                final File ondemandDir = new File(cpufreqDir, "ondemand");
                if(!ondemandDir.isDirectory())
                    {
                    errLogger.log("Cannot access "+ondemandDir+" dir.");
                    continue; // Try to continue for other CPUs...
                    }
                // Turn off the power-saving bias, as I think that it is obsolete and unhelpful.
                try { FileTools.writeTextFileInPlace(new File(ondemandDir, "powersave_bias"), "0"); }
                catch(final IOException e) { errLogger.log("Could not set powersave_bias flag for "+cpu+": " + e.getMessage()); }
                // Ignore niced work for speed selection iff conserving power.
                try { FileTools.writeTextFileInPlace(new File(ondemandDir, "ignore_nice_load"), lowPower ? "1" : "0"); }
                catch(final IOException e) { errLogger.log("Could not set ignore_nice_load flag for "+cpu+": " + e.getMessage()); }
                // Adjust the usage threshold % to turn up the frequency (default 80).
                try { FileTools.writeTextFileInPlace(new File(ondemandDir, "up_threshold"), lowPower ? "85" : "70"); }
                catch(final IOException e) { errLogger.log("Could not set up_threshold flag for "+cpu+": " + e.getMessage()); }
                // Adjust the sampling rate to be as fast as possible in full-power mode, else as slow as possible.
                // Note that the "sampling_rate_max" value is actually the maximum *interval* and thus minimum rate, and vv for _min.
                try { FileTools.writeTextFileInPlace(new File(ondemandDir, "sampling_rate"), FileTools.readTextFileLines(new File(ondemandDir, lowPower ? "sampling_rate_max" : "sampling_rate_min")).get(0)); }
                catch(final Exception e) { errLogger.log("Could not set sampling_rate flag for "+cpu+": " + e.getMessage()); }
                }
            }

        // Try to change the power-sensitive scheduling behaviour as appropriate.
        try { FileTools.writeTextFileInPlace(SYSFS_sched_mc_power_savings, lowPower ? "1" : "0"); }
        catch(final IOException e) { errLogger.log("Failed to update mc_power_savings mode: " + e.getMessage()); }

        // Try to create/remove the low-power output flag file if possible.
        if(outputFlagFile != null)
            {
            if(!lowPower)
                {
                // Not low power, so remove output flag file if extant.
                if(outputFlagFile.exists() && !outputFlagFile.delete()) { errLogger.log("Cannot remove low-power output flag file: " + outputFlagFile); }
                }
            else
                {
                // In low-power mode, so create output flag file if absent.
                if(!outputFlagFile.exists())
                    {
                    try
                        {
                        outputFlagFile.createNewFile();
                        if(!outputFlagFile.exists())  { errLogger.log("Cannot create low-power output flag file: " + outputFlagFile); }
                        }
                    catch(final IOException e) { errLogger.log("Error creating low-power output flag file: " + outputFlagFile + ": " + e.getMessage()); }
                    }
                }
            }
        }

    /**True to generate debug info for this run. */
    final boolean debug = false;

    /**Default/preferred CPU frequency governor name; never null.
     * The Intel/PowerTOP people claim that ondemand probably gives
     * best overall system performance and lowest power consumption.
     */
    public static final String DEFAULT_CPU_GOV = "ondemand";

    /**CPU governor that we will apply for all processors; never null.
     * This does not change during one run of this program.
     */
    public final String freqGovernor = DEFAULT_CPU_GOV;

    /**Default (immutable) set of flag files whose presense indicates low external power; never null.
     * If any of the names files exists,
     * this indicates that the low-power mode should be used.
     * <p>
     * To avoid confusion, any files listed should be absolute paths,
     * and the files should be plain files or possibly directories.
     */
    public static final Set<File> DEFAULT_LOW_POWER_FLAG_FILES = Collections.singleton(
        new File("/var/run/EXTERNAL_BATTERY_LOW.flag"));

    /**Current (immutable) set of low-power flag files for this run; never null. */
    public final Set<File> lowPowerFlagFiles = DEFAULT_LOW_POWER_FLAG_FILES;

    /**Old-style /proc file to parse for AC/battery power status; never null.
     * Normal format:
<pre>
state:                   on-line
</pre>
     */
    private static final File PROC_AC_POWER_STATUS_FILE = new File("/proc/acpi/ac_adapter/AC0/state");

//    /**Old-style /proc file to parse for CPU temperature; never null.
//     */
//    private static final File PROC_THRM_STATUS_FILE = new File("/proc/acpi/thermal_zone/THRM/temperature");

    /**Default minimum poll time for significant system power status changes (seconds); strictly positive,
     * Too small a time may waste energy; too large a time makes this unresponsive.
     * <p>
     * Realistically nothing of significance takes less than a few seconds,
     * and no reaction need take place in less than a few seconds,
     * so a value of at least 5 seconds is probably reasonable.
     */
    private static final int DEFAULT_POLL_S = 10;

    /**Current (immutable) minimum poll time for system power status changes; strictly positive.
     * Note that we may poll less frequently when in low-power mode
     * (ie the gap between polls may be significantly longer than this)
     * to reduce polling effort and to be conservative about leaving low-power mode.
     */
    public final int poll_s = DEFAULT_POLL_S;
    }
